/*
 * lxlist-i.i.c
 * Copyright (C) 2011-2019, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <fd.h>
#include <it.h>
#include <l_list.h>
#include <lxcall.h>
#include <lxcast.h>
#include <lxlist-drive.h>
#include <lxlist-inter.h>
#include <lxlist-types.h>

#define dxlist(list) \
    ((struct dxlist_type *) (list))
#define lxlist(list) \
    ((struct lxlist_type *) (list))

typedef struct roll_type {
    struct excase_type *excase_data;
    struct lxtext_type *lxtext_data;
    unsigned miss, size;
} roll_type;

typedef struct rule_type {
    int *args, roll;
    struct excase_type *excase_data;
    struct fxdata_type *fxdata_data;
    struct lxtext_type *lxtext_data;
} rule_type;

extern const void *const x1f4_a1_walk_link;

static int ever_list(struct lxlist_type *, struct fxdata_type *, int *,
		     struct dxlist_type *, int, int, int, unsigned *);
static int rule_list(void *, void *);
static int slip_roll(void *);

static int
ever_list(struct lxlist_type *lxlist_data, struct fxdata_type *fxdata_data,
	  int *post_args, struct dxlist_type *dxlist_data, int head, int tail,
	  int infer, unsigned *miss)
{
    int status;
    struct rule_type rule;

    rule.args = post_args;
    rule.excase_data = lock_excase(dxlist_data);
    rule.fxdata_data = fxdata_data;
    rule.lxtext_data = lxlist_data->link_f.text;

    rule.roll = 0;

    if (tail < head + infer) {
	unsigned size;

	size = dxlist_data->size;
	if (head < size) {
	    status = _libx1f4i0_over_l4list
		(dxlist_data->list, head, size - 1, &rule, rule_list);
	} else {
	    status = 0;
	}
	if (status) {
	} else {
	    if (tail) {
		status = _libx1f4i0_over_l4list
		    (dxlist_data->list, 0, tail - 1, &rule, rule_list);
	    }
	}
    } else {
	if (head ^ tail) {
	    status = _libx1f4i0_over_l4list
		(dxlist_data->list, head, tail - 1, &rule, rule_list);
	} else {
	    status = 0;
	}
    }

    if (rule.roll) {
	*miss = rule.fxdata_data - fxdata_data;
    }

    return status;
}


static int
rule_list(void *line, void *data)
{
    int link, status;
    struct rule_type *rule_data;

    rule_data = line;

    link = effect_link(data);

    *rule_data->args++ = effect_type(data);
    *rule_data->args++ = link;

    if (~link) {
	struct lxtext_type *lxtext_data;

	lxtext_data = rule_data->lxtext_data + link;

	if (lxtext_data->lxtype.flags & X1f4_LX_LINK_ACCESS) {
	    status = lxtext_data->lxtype.node
		(lxtext_data->lxtype.context, (void *) rule_data->fxdata_data,
		 E_USER(data), rule_data->excase_data);
	    if (status) {
	    } else {
		rule_data->roll = 1;
		rule_data->fxdata_data++;
	    }
	} else {
	    int (*copy)(void *, void *, void **, const void *,
			const struct x1f4_nodelink_type *);

	    copy = lxtext_data->lxtype.copy;
	    if (copy) {
		status = copy
		    (lxtext_data->lxtype.context, NULL,
		     (void *) rule_data->fxdata_data, E_USER(data),
		     rule_data->excase_data);
		if (status) {
		} else {
		    rule_data->roll = 1;
		    rule_data->fxdata_data++;
		}
	    } else {
		status = 0;

		*rule_data->fxdata_data++ =
		    *(struct fxdata_type *) (effect_data(data));
	    }
	}
    } else {
	status = 0;

	*rule_data->fxdata_data++ =
	    *(struct fxdata_type *) (effect_data(data));
    }

    return status;
}


static int
slip_roll(void *roll)
{
    int *args, status = 0;
    struct excase_type *excase_data;
    struct fxdata_type *fxdata_data;
    struct lxtext_type *lxtext_data;
    struct roll_type *roll_data;
    unsigned miss;

    roll_data = roll;

    fxdata_data = (void *) (roll_data + 1);

    excase_data = roll_data->excase_data;
    lxtext_data = roll_data->lxtext_data;

    args = (void *) (fxdata_data + roll_data->size);
    args++;

    miss = roll_data->miss;
    for (; miss; miss--) {
	int link;

	link = *args;
	if (~link) {
	    int excess, (*free)(void *, void *, struct excase_type **,
				const struct excase_type *);
	    struct lxtext_type *lxtext_slip;

	    lxtext_slip = lxtext_data + link;

	    free = lxtext_slip->lxtype.free;
	    if (free) {
		excess = free
		    (lxtext_slip->lxtype.context, I_USER(fxdata_data), NULL,
		     excase_data);
		if (excess) {
		    status = _libx1f4i0_lxcall_land_slip(excess, status);
		}
	    }
	}

	args += 2;

	fxdata_data++;
    }

    return status;
}


int
_libx1f4i0_lxlist_roll_list(void *lxlist, void *output, void **input,
			    void **track,
			    struct x1f4_linetext_type *linetext_data,
			    unsigned count,
			    const struct x1f4_function_type *function_text,
			    int lead, int lock, void *dxlist, int head,
			    int tail, int infer)
{
    int status;
    void *beta, **mind;

    status = lxlist(lxlist)->link_v.pick
	(lxlist(lxlist)->link_v.data, &mind, slip_roll,
	 sizeof(struct roll_type)
	 + count * (sizeof(struct fxdata_type) + (sizeof(int) << 1)), &beta);
    if (status) {
	status = _libx1f4i0_lxlist_stat_link(lxlist);
    } else {
	int *press;
	struct fxdata_type *fxdata_data;
	struct roll_type *roll_data;

	*mind = beta;

	roll_data = beta;

	fxdata_data = (void *) (roll_data + 1);

	press = (void *) (fxdata_data + count);

	roll_data->miss = 0;
	roll_data->size = count;

	roll_data->excase_data = lock_excase(dxlist);
	roll_data->lxtext_data = lxlist(lxlist)->link_f.text;

	status = ever_list
	    (lxlist, fxdata_data, press, dxlist, head, tail, infer,
	     &roll_data->miss);
	if (status) {
	} else {
	    if (output == x1f4_a1_walk_link) {
		status = _libx1f4i0_lxlist_lift_call
		    (lxlist, output, input, linetext_data, count,
		     function_text, lead, lock, press, fxdata_data);
	    } else {
		status = _libx1f4i0_lxlist_fold_call
		    (lxlist, output, input, linetext_data, count,
		     function_text, lead, lock, press, fxdata_data);
	    }
	}
    }

    return status;
}
